(*
 * This file is part of Bolt.
 * Copyright (C) 2009 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


open Camlp4

let level = ref 5

let logger = ref ""

let for_pack = ref ""

let logger_name modname =
  if !logger <> "" then
    !logger
  else if !for_pack <> "" then
    !for_pack ^ "." ^ modname
  else
    modname

module Make (Syntax : Sig.Camlp4Syntax) = struct
  open Sig
  include Syntax

  let level_of_string x = (* defined here to avoid dependency *)
    let _loc = Loc.ghost in
    match String.uppercase x with
    | "FATAL" -> <:expr< Bolt.Level.FATAL >>, 0
    | "ERROR" -> <:expr< Bolt.Level.ERROR >>, 1
    | "WARN" -> <:expr< Bolt.Level.WARN >>, 2
    | "INFO" -> <:expr< Bolt.Level.INFO >>, 3
    | "DEBUG" -> <:expr< Bolt.Level.DEBUG >>, 4
    | "TRACE" -> <:expr< Bolt.Level.TRACE >>, 5
    | _ -> failwith (Printf.sprintf "invalid logging level '%s'" x)

  type attributes = {
      error : Ast.expr option;
      properties : Ast.expr option;
      name : string option;
    }

  let no_attribute = { error = None; properties = None; name = None; }

  let merge x y =
    let m n x y = match x, y with
    | Some _, Some _ -> failwith (n ^ " attribute defined twice")
    | Some _, _ -> x
    | _, Some _ -> y
    | _ -> None in
    { error = m "'exception'" x.error y.error;
      properties = m "'properties'" x.properties y.properties;
      name = m "'name'" x.name y.name; }

  EXTEND Gram
    GLOBAL: expr;
    attr: [[ "EXCEPTION"; e = expr -> { no_attribute with error = Some e }
           | "NAME"; s = STRING -> { no_attribute with name = Some s }
           | "PROPERTIES"; e = expr -> { no_attribute with properties = Some e } ]];
    expr: LEVEL "simple" [[ "LOG"; msg = STRING; e = OPT [ expr ]; l = LIST0 attr; "LEVEL"; lvl = UIDENT ->
          let attrs = List.fold_left merge no_attribute l in
          let level_code, level_value = level_of_string lvl in
          if level_value <= !level then begin
            let file = Loc.file_name _loc in
            let basename = Filename.basename file in
            let logger = match attrs.name with
            | Some s -> s
            | None -> logger_name (String.capitalize (try Filename.chop_extension basename with _ -> basename)) in
            let line = string_of_int (Loc.start_line _loc) in
            let column = string_of_int ((Loc.start_off _loc) - (Loc.start_bol _loc)) in
            let properties = match attrs.properties with
            | Some x -> x
            | None -> <:expr< [] >> in
            let error = match attrs.error with
            | Some x -> <:expr< Some $x$ >>
            | None -> <:expr< None >> in
            (match e with
            | Some args ->
		let rec insert top x = function
		  | Ast.ExApp (loc, e1, e2) ->
		      (match e1 with
		      | Ast.ExApp _ -> insert false x e1
		      | _ ->
			  let node = Ast.ExApp (loc, x, e1) in
			  Ast.ExApp (loc, node, e2))
		  | v -> if top then Ast.ExApp (Ast.loc_of_expr v, x, v) else v in
		let msg = insert true (<:expr< Printf.sprintf $str:msg$ >>) args in
                <:expr< Bolt.Logger.log
                  $str:logger$
                  $level_code$
                  ~file:$str:file$
                  ~line:$int:line$
                  ~column:$int:column$
                  ~properties:$properties$
                  ~error:$error$
                  ($msg$) >>
            | None ->
                <:expr< Bolt.Logger.log
                  $str:logger$
                  $level_code$
                  ~file:$str:file$
                  ~line:$int:line$
                  ~column:$int:column$
                  ~properties:$properties$
                  ~error:$error$
                  $str:msg$ >>)
          end else
            <:expr< () >>
      ]];
  END
end

let () =
  let module Id = struct let name = "Bolt" let version = "1.0" end in
  let module M = Register.OCamlSyntaxExtension (Id) (Make) in
  Options.add
    "-level"
    (Arg.String
       (fun s -> match String.uppercase s with
       | "NONE" -> level := -1
       | "FATAL" -> level := 0
       | "ERROR" -> level := 1
       | "WARN" -> level := 2
       | "INFO" -> level := 3
       | "DEBUG" -> level := 4
       | "TRACE" -> level := 5
       | _ -> raise (Arg.Bad "invalid level")))
    "<level>  Set the logging level";
  Options.add "-logger" (Arg.Set_string logger) "<name>  Set the logger name";
  Options.add "-for-pack" (Arg.Set_string for_pack) "<prefix>  Set the prefix for logger names"
