(*
 * This file is part of Bolt.
 * Copyright (C) 2009 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module defines the events and how to construct them. *)


type time
(** The type of event timestamps. *)

type t = private {
    id : int; (** Event identifier. *)
    thread : int; (** Threaf identifier of event source. *)
    timestamp : time; (** Event timestamp. *)
    relative : int; (** Time elapsed between application start and event construction. *)
    level : Level.t; (** Event level. *)
    logger : string; (** Logger name. *)
    file : string; (** Location of event source. *)
    line : int; (** Location of event source. *)
    column : int; (** Location of event source. *)
    message : string; (** Event message. *)
    properties : (string * string) list; (** Event properties as an association list. *)
    error : (exn * string) option; (** Event error (parameters being actual exception and exception backtrace). *)
  }
(** The type of log events. *)

val make : string -> Level.t -> ?file:string -> ?line:int -> ?column:int -> ?properties:(string * string) list -> ?error:exn option -> string -> t
(** [make lg lv ~file:fn ~line:ln ~column:cl ~properties:p ~error:e m]
    constructs a log event for logger [lg] with level [lv], location defined by
    [fn] (filename), [ln] (line) and [cl] (column). [p] is an association list
    providing user-defined bindings, and [e] is an optional exception to be
    recorded. Identifier, thread, timestamp, relative time and backtrace are
    automatically generated. *)

val with_logger : string -> t -> t
(** [with_logger l e] returns an event that is identical to [e], except that
    logger is equal to [l]. *)

val bindings : t -> (string * string) list
(** Returns the bindings for the passed event.
    The bindings are an association list to be used for event rendering.
    The followinf keys are defined:
    - ["id"] event identifier;
    - ["thread"] thread identifier;
    - ["sec"] seconds of event timestamp;
    - ["min"] minutes of event timestamp;
    - ["hour"] hour of event timestamp;
    - ["mday"] day of month of event timestamp;
    - ["month"] month of year of event timestamp;
    - ["year"] year of event timestamp;
    - ["wday"] day of week of event timestamp;
    - ["time"] event timestamp;
    - ["relative"] time elapsed between initilization and event creation;
    - ["level"] event level;
    - ["logger"] event logger;
    - ["file"] event file;
    - ["filebase"] event file (without directory information);
    - ["line"] event line;
    - ["column"] event column;
    - ["message"] event message;
    - ["properties"] property list of event (format: ["[k1: v1; ...; kn: vn]"]);
    - ["exception"] event exception;
    - ["backtrace"] event exception backtrace. *)

(**/**)

val register_get_thread_id : (unit -> int) -> unit
(** [register_get_thread_id f] register [f] to be the function returning the
    identifier of the current thread. *)
