(*
 * This file is part of Bolt.
 * Copyright (C) 2009 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


(* Definitions *)

type t = (string list) * (string list) * (Event.t -> string)

let layouts = Hashtbl.create 17

let register name layout =
  Hashtbl.replace layouts name layout

let register_unnamed layout =
  let i = ref 0 in
  while Hashtbl.mem layouts (string_of_int !i) do
    incr i
  done;
  let name = string_of_int !i in
  register name layout;
  name

let get name =
  Hashtbl.find layouts name


(* Predefined layouts *)

let decode s =
  try
    let idx = String.index s ':' in
    let len = String.length s in
    let key = String.sub s 0 idx in
    let size = String.sub s (succ idx) (len - idx - 1) in
    key, (try int_of_string size with _ -> 0)
  with Not_found -> s, 0

let assoc l s =
  try
    let key, size = decode s in
    let res = List.assoc key l in
    let left = size < 0 in
    let pad_size = (abs size) - (String.length res) in
    let pad_string = String.make (max 0 pad_size) ' ' in
    if left then pad_string ^ res else res ^ pad_string
  with _ ->
    ""

let render_bindings b fmt =
  let buffer = Buffer.create 32 in
  try
    Buffer.add_substitute buffer (assoc b) fmt;
    Buffer.contents buffer
  with _ -> fmt

let render fmt e =
  render_bindings (Event.bindings e) fmt
  
let simple =
  [], [], render "$(level:5) - $(message)"

let default =
  [], [], render "$(relative:-6) [$(file:-16) $(line:4)] $(level:5) - $(message)"

let pattern h f p =
  h, f, render p

let style = [
  "<style type=\"text/css\">" ;
  "table.sample {" ;
  "  border-width: 1px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  border-collapse: collapse;" ;
  "  background-color: white;" ;
  "}" ;
  "table.sample th {" ;
  "  border-width: 1px;" ;
  "  padding: 3px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  background-color: white;" ;
  "}" ;
  "table.sample td {" ;
  "  border-width: 1px;" ;
  "  padding: 3px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  background-color: white;" ;
  "}" ;
  "</style>"
]

let html =
  let version = "1.0" in
  [ "<html>" ;
    "<head>" ;
    "<title>Log</title>" ] @
  style @
  [ "</head>" ;
    "<body>" ;
    "<p align=\"center\">" ;
    "<table class=\"sample\">" ;
    ("<tr>" ^
     "<th>Identifier</th>" ^
     "<th>Time</th>" ^
     "<th>Millis</th>" ^
     "<th>Level</th>" ^
     "<th>File</th>" ^
     "<th>Message</th>" ^
     "</tr>") ],
  [ "</table>" ;
    "</p>" ;
    ("<p style=\"font-size: smaller; text-align: center;\">Generated by <a href=\"http://bolt.x9c.fr\">Bolt " ^ version ^ "</a></p>") ;
    "</body>" ;
    "</html>" ],
  render ("<tr style=\"font-family: monospace;\">" ^
          "<td align=\"right\">$(id)</td>" ^
          "<td>$(hour):$(min):$(sec)</td>" ^
          "<td align=\"right\">$(relative)</td>" ^
          "<td>$(level)</td>" ^
          "<td><pre>$(filebase:-24) $(line:4)</pre></td>" ^
          "<td>$(message)</td>" ^
          "</tr>")

let protect_cdata s =
  let len = String.length s in
  let buf = Buffer.create len in
  let i = ref 0 in
  while (!i < len) do
    if (!i >= 2) && (s.[!i - 2] = ']') && (s.[!i - 1] = ']') && (s.[!i] = '>') then
      Buffer.add_string buf " >"
    else
      Buffer.add_char buf s.[!i];
    incr i
  done;
  Buffer.contents buf

let xml =
  [],
  [],
  (fun e ->
    let properties =
      List.map
        (fun (k, v) -> Printf.sprintf "  <log4j:data name=\"%s\" value=\"%s\"/>\n" k v)
        e.Event.properties in
    render_bindings
      (List.map
	 (fun (k, v) -> (k, (if k = "message" then protect_cdata v else v)))
	 (Event.bindings e))
      ("<log4j:event logger=\"$(logger)\" level=\"$(level)\" thread=\"$(thread)\" timestamp=\"$(time)\">\n" ^
       "<log4j:message><![CDATA[$(message)]]></log4j:message>\n" ^
       (match e.Event.error with
       | Some (_, s) -> "<log4j:throwable><![CDATA[" ^ (protect_cdata s) ^ "]]></log4j:throwable>\n"
       | None -> "") ^
       "<log4j:locationInfo class=\"Unknown\" method=\"unknown()\" file=\"$(file)\" line=\"$(line)\"/>\n" ^
       "<log4j:properties>\n" ^
       (String.concat "" properties) ^
       "</log4j:properties>\n" ^
       "</log4j:event>\n"))

let csv sep l =
  let render e =
    let b = Event.bindings e in
    let values = List.map (fun k -> try List.assoc k b with _ -> "") l in
    String.concat sep values in
  ([], [], render)

let () =
  List.iter
    (fun (x, y) -> register x y)
    [ "simple",  simple ;
      "default", default ;
      "html",    html ;
      "xml",     xml ]
